@@ -26,18 +26,18 @@
  * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *************************************************************************/
 
-// include flann_cpp stuff
-#include "flann_cpp.cpp"
-
 #define FLANN_FIRST_MATCH
 
+#include "flann.h"
+
 
 struct FLANNParameters DEFAULT_FLANN_PARAMETERS = {
     FLANN_INDEX_KDTREE,
-    32, 0.2, 0.0,
+    32, 0.0f,
+    0, -1, 0,
     4, 4,
-    32, 11, FLANN_CENTERS_RANDOM,
-    0.9, 0.01, 0, 0.1,
+    32, 11, FLANN_CENTERS_RANDOM, 0.2f,
+    0.9f, 0.01f, 0, 0.1f,
     FLANN_LOG_NONE, 0
 };
 
@@ -45,6 +45,126 @@ struct FLANNParameters DEFAULT_FLANN_PARAMETERS = {
 using namespace flann;
 
 
+flann::IndexParams create_parameters(FLANNParameters* p)
+{
+    flann::IndexParams params;
+
+    params["algorithm"] = p->algorithm;
+
+    params["checks"] = p->checks;
+    params["cb_index"] = p->cb_index;
+    params["eps"] = p->eps;
+
+    if (p->algorithm == FLANN_INDEX_KDTREE) {
+        params["trees"] = p->trees;
+    }
+
+    if (p->algorithm == FLANN_INDEX_KDTREE_SINGLE) {
+        params["trees"] = p->trees;
+        params["leaf_max_size"] = p->leaf_max_size;
+    }
+
+#ifdef FLANN_USE_CUDA
+    if (p->algorithm == FLANN_INDEX_KDTREE_CUDA) {
+        params["leaf_max_size"] = p->leaf_max_size;
+    }
+#endif
+
+    if (p->algorithm == FLANN_INDEX_KMEANS) {
+        params["branching"] = p->branching;
+        params["iterations"] = p->iterations;
+        params["centers_init"] = p->centers_init;
+    }
+
+    if (p->algorithm == FLANN_INDEX_AUTOTUNED) {
+        params["target_precision"] = p->target_precision;
+        params["build_weight"] = p->build_weight;
+        params["memory_weight"] = p->memory_weight;
+        params["sample_fraction"] = p->sample_fraction;
+    }
+
+    if (p->algorithm == FLANN_INDEX_HIERARCHICAL) {
+        params["branching"] = p->branching;
+        params["centers_init"] = p->centers_init;
+        params["trees"] = p->trees;
+        params["leaf_max_size"] = p->leaf_max_size;
+    }
+
+    if (p->algorithm == FLANN_INDEX_LSH) {
+        params["table_number"] = p->table_number_;
+        params["key_size"] = p->key_size_;
+        params["multi_probe_level"] = p->multi_probe_level_;
+    }
+
+    params["log_level"] = p->log_level;
+    params["random_seed"] = p->random_seed;
+
+    return params;
+}
+
+flann::SearchParams create_search_params(FLANNParameters* p)
+{
+    flann::SearchParams params;
+    params.checks = p->checks;
+    params.eps = p->eps;
+    params.sorted = p->sorted;
+    params.max_neighbors = p->max_neighbors;
+    params.cores = p->cores;
+
+    return params;
+}
+
+
+void update_flann_parameters(const IndexParams& params, FLANNParameters* flann_params)
+{
+	if (has_param(params,"algorithm")) {
+		flann_params->algorithm = get_param<flann_algorithm_t>(params,"algorithm");
+	}
+	if (has_param(params,"trees")) {
+		flann_params->trees = get_param<int>(params,"trees");
+	}
+	if (has_param(params,"leaf_max_size")) {
+		flann_params->leaf_max_size = get_param<int>(params,"leaf_max_size");
+	}
+	if (has_param(params,"branching")) {
+		flann_params->branching = get_param<int>(params,"branching");
+	}
+	if (has_param(params,"iterations")) {
+		flann_params->iterations = get_param<int>(params,"iterations");
+	}
+	if (has_param(params,"centers_init")) {
+		flann_params->centers_init = get_param<flann_centers_init_t>(params,"centers_init");
+	}
+	if (has_param(params,"target_precision")) {
+		flann_params->target_precision = get_param<float>(params,"target_precision");
+	}
+	if (has_param(params,"build_weight")) {
+		flann_params->build_weight = get_param<float>(params,"build_weight");
+	}
+	if (has_param(params,"memory_weight")) {
+		flann_params->memory_weight = get_param<float>(params,"memory_weight");
+	}
+	if (has_param(params,"sample_fraction")) {
+		flann_params->sample_fraction = get_param<float>(params,"sample_fraction");
+	}
+	if (has_param(params,"table_number")) {
+		flann_params->table_number_ = get_param<unsigned int>(params,"table_number");
+	}
+	if (has_param(params,"key_size")) {
+		flann_params->key_size_ = get_param<unsigned int>(params,"key_size");
+	}
+	if (has_param(params,"multi_probe_level")) {
+		flann_params->multi_probe_level_ = get_param<unsigned int>(params,"multi_probe_level");
+	}
+	if (has_param(params,"log_level")) {
+		flann_params->log_level = get_param<flann_log_level_t>(params,"log_level");
+	}
+	if (has_param(params,"random_seed")) {
+		flann_params->random_seed = get_param<long>(params,"random_seed");
+	}
+}
+
+
 void init_flann_parameters(FLANNParameters* p)
 {
     if (p != NULL) {
@@ -82,23 +202,24 @@ flann_index_t __flann_build_index(typename Distance::ElementType* dataset, int r
         if (flann_params == NULL) {
             throw FLANNException("The flann_params argument must be non-null");
         }
-        IndexParams* params = IndexParams::createFromParameters(*flann_params);
-        Index<Distance>* index = new Index<Distance>(Matrix<ElementType>(dataset,rows,cols), *params, d);
+        IndexParams params = create_parameters(flann_params);
+        Index<Distance>* index = new Index<Distance>(Matrix<ElementType>(dataset,rows,cols), params, d);
         index->buildIndex();
-        const IndexParams* index_params = index->getIndexParameters();
-        index_params->toParameters(*flann_params);
 
-        if (index->getIndex()->getType()==FLANN_INDEX_AUTOTUNED) {
-            AutotunedIndex<Distance>* autotuned_index = (AutotunedIndex<Distance>*)index->getIndex();
-            flann_params->checks = autotuned_index->getSearchParameters()->checks;
-            *speedup = autotuned_index->getSpeedup();
+        if (flann_params->algorithm==FLANN_INDEX_AUTOTUNED) {
+            IndexParams params = index->getParameters();
+            update_flann_parameters(params,flann_params);
+            SearchParams search_params = get_param<SearchParams>(params,"search_params");
+            *speedup = get_param<float>(params,"speedup");
+            flann_params->checks = search_params.checks;
+            flann_params->eps = search_params.eps;
+            flann_params->cb_index = get_param<float>(params,"cb_index",0.0);
         }
 
-        delete params;
         return index;
     }
     catch (std::runtime_error& e) {
-        logger.error("Caught exception: %s\n",e.what());
+        Logger::error("Caught exception: %s\n",e.what());
         return NULL;
     }
 }
@@ -128,7 +249,7 @@ flann_index_t _flann_build_index(T* dataset, int rows, int cols, float* speedup,
         return __flann_build_index<KL_Divergence<T> >(dataset, rows, cols, speedup, flann_params);
     }
     else {
-        logger.error( "Distance type unsupported in the C bindings, use the C++ bindings instead\n");
+        Logger::error( "Distance type unsupported in the C bindings, use the C++ bindings instead\n");
         return NULL;
     }
 }
@@ -172,7 +293,7 @@ int __flann_save_index(flann_index_t index_ptr, char* filename)
         return 0;
     }
     catch (std::runtime_error& e) {
-        logger.error("Caught exception: %s\n",e.what());
+        Logger::error("Caught exception: %s\n",e.what());
         return -1;
     }
 }
@@ -202,7 +323,7 @@ int _flann_save_index(flann_index_t index_ptr, char* filename)
         return __flann_save_index<KL_Divergence<T> >(index_ptr, filename);
     }
     else {
-        logger.error( "Distance type unsupported in the C bindings, use the C++ bindings instead\n");
+        Logger::error( "Distance type unsupported in the C bindings, use the C++ bindings instead\n");
         return -1;
     }
 }
@@ -242,7 +363,7 @@ flann_index_t __flann_load_index(char* filename, typename Distance::ElementType*
         return index;
     }
     catch (std::runtime_error& e) {
-        logger.error("Caught exception: %s\n",e.what());
+        Logger::error("Caught exception: %s\n",e.what());
         return NULL;
     }
 }
@@ -272,7 +393,7 @@ flann_index_t _flann_load_index(char* filename, T* dataset, int rows, int cols)
         return __flann_load_index<KL_Divergence<T> >(filename, dataset, rows, cols);
     }
     else {
-        logger.error( "Distance type unsupported in the C bindings, use the C++ bindings instead\n");
+        Logger::error( "Distance type unsupported in the C bindings, use the C++ bindings instead\n");
         return NULL;
     }
 }
@@ -314,20 +435,20 @@ int __flann_find_nearest_neighbors(typename Distance::ElementType* dataset,  int
     try {
         init_flann_parameters(flann_params);
 
-        IndexParams* params = IndexParams::createFromParameters(*flann_params);
-        Index<Distance>* index = new Index<Distance>(Matrix<ElementType>(dataset,rows,cols), *params, d);
+        IndexParams params = create_parameters(flann_params);
+        Index<Distance>* index = new Index<Distance>(Matrix<ElementType>(dataset,rows,cols), params, d);
         index->buildIndex();
         Matrix<int> m_indices(result,tcount, nn);
         Matrix<DistanceType> m_dists(dists,tcount, nn);
+        SearchParams search_params = create_search_params(flann_params);
         index->knnSearch(Matrix<ElementType>(testset, tcount, index->veclen()),
                          m_indices,
-                         m_dists, nn, SearchParams(flann_params->checks) );
+                         m_dists, nn, search_params );
         delete index;
-        delete params;
         return 0;
     }
     catch (std::runtime_error& e) {
-        logger.error("Caught exception: %s\n",e.what());
+        Logger::error("Caught exception: %s\n",e.what());
         return -1;
     }
 
@@ -360,7 +481,7 @@ int _flann_find_nearest_neighbors(T* dataset,  int rows, int cols, T* testset, i
         return __flann_find_nearest_neighbors<KL_Divergence<T> >(dataset, rows, cols, testset, tcount, result, dists, nn, flann_params);
     }
     else {
-        logger.error( "Distance type unsupported in the C bindings, use the C++ bindings instead\n");
+        Logger::error( "Distance type unsupported in the C bindings, use the C++ bindings instead\n");
         return -1;
     }
 }
@@ -408,14 +529,15 @@ int __flann_find_nearest_neighbors_index(flann_index_t index_ptr, typename Dista
         Matrix<int> m_indices(result,tcount, nn);
         Matrix<DistanceType> m_dists(dists, tcount, nn);
 
+        SearchParams search_params = create_search_params(flann_params);
         index->knnSearch(Matrix<ElementType>(testset, tcount, index->veclen()),
                          m_indices,
-                         m_dists, nn, SearchParams(flann_params->checks) );
+                         m_dists, nn, search_params );
 
         return 0;
     }
     catch (std::runtime_error& e) {
-        logger.error("Caught exception: %s\n",e.what());
+        Logger::error("Caught exception: %s\n",e.what());
         return -1;
     }
 
@@ -448,7 +570,7 @@ int _flann_find_nearest_neighbors_index(flann_index_t index_ptr, T* testset, int
         return __flann_find_nearest_neighbors_index<KL_Divergence<T> >(index_ptr, testset, tcount, result, dists, nn, flann_params);
     }
     else {
-        logger.error( "Distance type unsupported in the C bindings, use the C++ bindings instead\n");
+        Logger::error( "Distance type unsupported in the C bindings, use the C++ bindings instead\n");
         return -1;
     }
 }
@@ -501,15 +623,16 @@ int __flann_radius_search(flann_index_t index_ptr,
 
         Matrix<int> m_indices(indices, 1, max_nn);
         Matrix<DistanceType> m_dists(dists, 1, max_nn);
+        SearchParams search_params = create_search_params(flann_params);
         int count = index->radiusSearch(Matrix<ElementType>(query, 1, index->veclen()),
                                         m_indices,
-                                        m_dists, radius, SearchParams(flann_params->checks) );
+                                        m_dists, radius, search_params );
 
 
         return count;
     }
     catch (std::runtime_error& e) {
-        logger.error("Caught exception: %s\n",e.what());
+        Logger::error("Caught exception: %s\n",e.what());
         return -1;
     }
 }
@@ -545,7 +668,7 @@ int _flann_radius_search(flann_index_t index_ptr,
         return __flann_radius_search<KL_Divergence<T> >(index_ptr, query, indices, dists, max_nn, radius, flann_params);
     }
     else {
-        logger.error( "Distance type unsupported in the C bindings, use the C++ bindings instead\n");
+        Logger::error( "Distance type unsupported in the C bindings, use the C++ bindings instead\n");
         return -1;
     }
 }
@@ -620,7 +743,7 @@ int __flann_free_index(flann_index_t index_ptr, FLANNParameters* flann_params)
         return 0;
     }
     catch (std::runtime_error& e) {
-        logger.error("Caught exception: %s\n",e.what());
+        Logger::error("Caught exception: %s\n",e.what());
         return -1;
     }
 }
@@ -650,7 +773,7 @@ int _flann_free_index(flann_index_t index_ptr, FLANNParameters* flann_params)
         return __flann_free_index<KL_Divergence<T> >(index_ptr, flann_params);
     }
     else {
-        logger.error( "Distance type unsupported in the C bindings, use the C++ bindings instead\n");
+        Logger::error( "Distance type unsupported in the C bindings, use the C++ bindings instead\n");
         return -1;
     }
 }
@@ -699,7 +822,7 @@ int __flann_compute_cluster_centers(typename Distance::ElementType* dataset, int
         return clusterNum;
     }
     catch (std::runtime_error& e) {
-        logger.error("Caught exception: %s\n",e.what());
+        Logger::error("Caught exception: %s\n",e.what());
         return -1;
     }
 }
@@ -730,7 +853,7 @@ int _flann_compute_cluster_centers(T* dataset, int rows, int cols, int clusters,
         return __flann_compute_cluster_centers<KL_Divergence<T> >(dataset, rows, cols, clusters, result, flann_params);
     }
     else {
-        logger.error( "Distance type unsupported in the C bindings, use the C++ bindings instead\n");
+        Logger::error( "Distance type unsupported in the C bindings, use the C++ bindings instead\n");
         return -1;
     }
 }
